# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ AwakenApp 接口应用程序 ]

"""
from flask import Flask
from flask.blueprints import Blueprint

from ..api.blueprint import __all__ as blueprint_all

from ...baseic.log import LOG
from ...baseic.config import CONFIG
from ...baseic.decorator import singleton_pattern


@singleton_pattern
class AwakenApp(object):
    """
    [ AwakenApp 接口应用程序 ]

    ---
    描述:
        基于 Flask 实现的 Flask.WebApi 应用程序。

    """
    
    _app_living: Flask = Flask(__name__)
    """ 应用程序实例 """


    def __init__(self):
        """ 
        [ AwakenApp 接口应用程序 ]
        
        """
        self.update_config({'JSON_AS_ASCII': False})

        for blueprint in blueprint_all:
            self.load_blueprint(blueprint)


    @property
    def living(self) -> Flask:
        """
        [ 应用程序实例 ]

        ---
        描述:
            返回应用程序实例。

        """
        return self._app_living


    def update_config(self, config_dict: dict) -> None:
        """
        [ 更新配置项 ]

        ---
        描述:
            更新 Flask 应用程序的配置项。

        """
        self._app_living.config.update(config_dict)

        if CONFIG.debug:
            for ck, cv in config_dict.items():
                LOG.info(f'{ self.__class__.__name__ } :: 更新配置 :: [{ ck } = { cv }]')


    def load_blueprint(self, blueprint: Blueprint):
        """
        [ 载入蓝图 ]

        ---
        描述:
            将蓝图对象载入 Flask 应用程序。

        ---
        参数:
            blueprint { Blueprint } : Flask.Blueprint 蓝图对象

        """
        self._app_living.register_blueprint(blueprint)

        if CONFIG.debug:
            LOG.info(f'{ self.__class__.__name__ } :: 载入蓝图 :: [{ blueprint }]')



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::运行任务
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from ..awaken_info import ProjectInfo
from ..baseic.const import CONST
from ..baseic.common import window_template_output
from ..core.perform.perform_pool import PerFormPool
from ..core.interpreter.grammar_parser import GrammarParser


def instruction_running_task(argv: list):
    """
    [ 命令行指令::运行任务 ]

    ---
    参数:
        argv { list } : 参数列表

    """
    task_list = []

    if len(argv) >= 1:

        if not CONST.Path.FilePath.Init.exists():
            window_template_output(
                f'Awaken2 - {ProjectInfo.Version}', 
                [
                    '- 运行任务 -',
                    None,
                    '当前运行路径不是工程根目录;',
                    '请在工程根目录执行该命令.'
                ]
            )
            exit(0)
        
        # 循环解析任务参数
        for task_arg in argv:

            if task_arg in ['.', '/', './']:
                task_arg_path = CONST.Path.CWD
            else:
                task_arg_path = CONST.Path.CWD.joinpath(task_arg)  

            # 如果参数路径指向目录
            if task_arg_path.is_dir():
                awaken_list = task_arg_path.glob(f'*.awaken-*')
                for awaken_file in awaken_list:
                    task_list.append(awaken_file)

            # 如果参数路径指向文件
            else:
                if task_arg_path.exists():
                    task_list.append(task_arg_path)

        if len(task_list) < 1:
            window_template_output(
                f'Awaken2 - {ProjectInfo.Version}', 
                [
                    '- 运行任务 -',
                    None,
                    '暂无任务可供执行 !',
                ]
            )
            exit(0)

        # 多进程执行任务
        window_template_output(
            f'Awaken2 - {ProjectInfo.Version}', 
            [
                '- 运行任务 -',
                None,
                '多进程执行器正在消耗任务...',
            ]
        )
        grammar_parser = GrammarParser()
        perform_pool = PerFormPool()
        for task in task_list:
            task = grammar_parser.parsing(task)
            perform_pool.put_task(task)
        perform_pool.running()

    else:
        window_template_output(
            f'Awaken2 - {ProjectInfo.Version}', 
            [
                '- 运行任务 -',
                None,
                '运行项目指令参数异常:',
                None,
                '示例:',
                '>> awaken -run task1',
                '>> awaken -run task1 task2',
                '>> awaken -run task1 task2 project1',
            ]
        )
        exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 服务端请求处理程序 ]

"""
import json
from typing import Union

from flask import jsonify
from flask import request
from flask import Response

from ...baseic.const import CONST
from ...baseic.error import AwakenWebApiServerError
from ...core.interpreter.common import converting_para_format


URL_PREFIX: str = '/api'
""" URL前缀 """


class ServerRequestHandler(object):
    """
    [ 服务端请求处理程序 ]
    
    ---
    描述:
        处理 WEBAPI 请求的相关事务, 如解析前端请求参数与返回响应对象。

    """


    @staticmethod
    def analysis_request_parameter(keys: list, must_keys: Union[list, None] = None) -> dict:
        """
        [ 解析请求参数 ]

        ---
        描述:
            解析请求中指定 Key, 如果该值为空, 则舍弃该 Key, 而不是返回 None。

            如果需要确保解析结果中必须携带指定 Key, 请通过 must_keys 定义。

        ---
        参数:
            keys { list } : 待解析的 Key 列表
            must_keys { Union[list, None], optional } : 强制携带的 Key 列表, 默认为空

        ---
        示例:
            >>> params = ServerParameterHandler.analysis_request_parameter(keys=['key1', 'key2'])
            :: 该示例表示解析请求中的 key1 与 key2, 如果请求中缺失 key2, 则只返回 key1。

            >>> params = ServerParameterHandler.analysis_request_parameter(keys=['key1', 'key2'], must_keys=['key1])
            :: 该示例表示获取请求中的 key1 与 key2, 并且确保 key1 必须存在, 否则将引发异常。

        ---
        异常:
            AwakenWebApiServerError : 接口服务器异常。

        ---
        返回:
            dict : 返回解析完成的参数字典。

        """
        request_handle = request.args
        if request.method == 'POST':
            try:
                request_handle = json.loads(request.data.decode('UTF-8'))
            except json.decoder.JSONDecodeError:
                request_handle = request.form

        return ServerRequestHandler._common_analysis_request_keys(request_handle, keys, must_keys)


    @staticmethod
    def analysis_request_headers(keys: list, must_keys: Union[list, None] = None) -> dict:
        """
        [ 解析请求头 ]

        """
        headers_handle = request.headers
        return ServerRequestHandler._common_analysis_request_keys(headers_handle, keys, must_keys)


    @staticmethod
    def successful(paras: dict) -> Response:
        """
        [ 请求成功 ]

        ---
        描述:
            将数据序列化为 Json 并包装成 Flask.Response 响应体返回。

        ---
        参数:
            paras { dict } : 需要返回的数据

        ---
        返回:
            Response : 响应体对象
         
        """
        return jsonify({
            'code': CONST.Api.MessageCode.Success, 
            'message': CONST.State.Result.Success, 
            'result': paras
        })


    @staticmethod
    def unsuccessful(error: str) -> Response:
        """
        [ 请求失败 ]

        ---
        描述:
            将数据序列化为 Json 并包装成 Flask.Response 响应体返回。

        ---
        参数:
            error { str } : 需要返回的错误内容

        ---
        返回:
            Response : 响应体对象
         
        """
        return jsonify({
            'code': CONST.Api.MessageCode.Error, 
            'message': CONST.State.Result.Error, 
            'error': error
        })


    @staticmethod
    def _common_analysis_request_keys(data: dict, keys: list, must_keys: list) -> dict:
        """
        [ 通用的解析参数 ]

        ---
        描述:
            解析请求参数与解析请求头的通用函数。

        ---
        参数:
            data { dict } : 待解析的数据
            keys { list } : 待解析的 Key 列表
            must_keys { list } : 强制携带的 Key 列表
        
        ---
        异常:
            AwakenWebApiServerError : 接口服务器异常

        """
        decryption_parameter: dict = {}

        # 如果 key 为空, 则丢弃, 否则储存
        for key in keys:
            value = data.get(key)
            if not value: 
                ...
            else:
                decryption_parameter.update({key: converting_para_format(value)})
        
        # 检查 keys 中是否包含所有 must_keys 的成员元素
        # 否则抛出 ServerAnalysisParameterError 异常
        if must_keys:
            
            for key in must_keys:
                if key not in decryption_parameter.keys():
                    raise AwakenWebApiServerError(f'请求数据不存在必要参数 {key} !')

        return decryption_parameter
            



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 引擎代理
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from .web.web_runner import WebRunner
from ..baseic.const import CONST
from ..core.interpreter.structural import AwakenTask
from ..core.interpreter.task_preprocessor import TaskPreprocessor


class EngineAgency: ...


class EngineAgency:
    """
    [ 引擎代理 ]

    ---
    描述:
        NULL

    """
    
    task_preprocessor: TaskPreprocessor
    """ 任务预处理程序 """


    def __init__(self) -> None:
        self.task_preprocessor = TaskPreprocessor()


    def start(self, task: AwakenTask):

        task = self.task_preprocessor.pretreatment(task)

        if task.task_type == CONST.Type.Task.Web:
            engine = ''
            runner = WebRunner(task)
            runner.start(engine)

        # if task.task_type == CONST.Type.Task.Api:
        #     engine = ApiEngine()
        #     runner = ApiRunner(task)
        #     runner.start(engine)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 任务相关SQL::用例相关 ]

"""
from ..common import get_desc_limit
from ..common import parse_case_search_criteria
from ....baseic.const import CONST
from ....baseic.keyword import KEYWORD


class SqlCase(object):
    """
    [ 任务相关SQL::用例相关 ]

    """


    @staticmethod
    def get_count(paras: dict = {}) -> str:
        """
        [ 查询用例数量 ]

        ---
        可选参数:
            tid           { int } : 任务编号
            type          { str } : 类型
            state         { str } : 状态
            created_date  { str } : 创建日期, 例如 2022-02-02

        """
        # 解析搜索条件
        where_string = parse_case_search_criteria(paras)

        return f"""

        SELECT COUNT({ KEYWORD.Api.Case.Cid }) 
        
        FROM { KEYWORD.DataBase.TableName.Cases }

        {where_string}

        """


    @staticmethod
    def get_id_list(paras: dict = {}) -> str:
        """
        [ 查询 ID 数量 ]

        ---
        可选参数:
            tid           { int } : 任务编号
            type          { str } : 类型
            state         { str } : 状态
            created_date  { str } : 创建日期, 例如 2022-02-02

        """
        # 解析搜索条件
        where_string = parse_case_search_criteria(paras)

        return f"""

        SELECT { KEYWORD.Api.Case.Cid }
        
        FROM { KEYWORD.DataBase.TableName.Cases }

        {where_string}

        ORDER BY { KEYWORD.Api.Case.Cid } DESC

        """

    @staticmethod
    def get_case_list(paras: dict = {}) -> str:
        """
        [ 查询任务所有用例详情信息列表 ]

        ---
        可选参数:
            tid           { int } : 所属任务编号
            cid           { int } : 用例编号
            type          { str } : 类型
            state         { str } : 状态
            created_date  { str } : 创建日期, 例如 2022-02-02
            page          { int } : 页数
            number        { int } : 单页数据量

        """
        # 解析搜索条件
        where_string = parse_case_search_criteria(paras)

        limit = ''
        if 'page' in paras.keys() and 'number' in paras.keys():
            limit = f"LIMIT { get_desc_limit(paras['page'], paras['number']) }"

        return f"""

        SELECT * 
        
        FROM { KEYWORD.DataBase.TableName.Cases }

        {where_string}

        ORDER BY { KEYWORD.Api.Case.Cid } DESC

        {limit}

        """


    @staticmethod
    def creation(
        tid: int,
        name: str,
        type: str,
        remark: str,
        created_date: str
    ) -> str:
        """
        [ 创建用例 ]

        ---
        参数:
            tid          { int } : 所属任务编号
            name         { str } : 用例名称
            type         { str } : 类型
            remark       { str } : 用例备注信息
            created_date { str } : 创建日期

        """
        return f""" 

        INSERT INTO { KEYWORD.DataBase.TableName.Cases } (

            { KEYWORD.Api.Case.Cid }, 
            { KEYWORD.Api.Case.Tid }, 
            { KEYWORD.Api.Case.Name },
            { KEYWORD.Api.Case.Type }, 
            { KEYWORD.Api.Case.Remark }, 
            { KEYWORD.Api.Case.State }, 
            { KEYWORD.Api.Case.CreatedDate }, 
            { KEYWORD.Api.Case.RunStep }
        )

        VALUES (

            NULL, 
            '{ tid }', 
            '{ name }',
            '{ type }',
            '{ remark }', 
            '{ CONST.State.Result.Null }', 
            '{ created_date }', 
             NULL
        )

        """


    @staticmethod
    def update(
        cid: int,
        state: str,
        start_time: str,
        end_time: str,
        spend_time: str,
        run_step: str
    ) -> str:
        """
        [ 更新用例 ]

        ---
        参数:
            cid        { int } : 用例编号
            state      { str } : 用例状态
            start_time { str } : 开始时间
            end_time   { str } : 结束时间
            spend_time { str } : 消耗时间
            run_step   { str } : 运行步骤

        """
        return f"""

        UPDATE { KEYWORD.DataBase.TableName.Cases }

        SET 
            { KEYWORD.Api.Case.State }     = "{ state }", 
            { KEYWORD.Api.Case.StartTime } = "{ start_time }", 
            { KEYWORD.Api.Case.EndTime }   = "{ end_time }", 
            { KEYWORD.Api.Case.SpendTime } = "{ spend_time }", 
            { KEYWORD.Api.Case.RunStep }   = "{ run_step }"

        WHERE
            { KEYWORD.Api.Case.Cid } = { cid }

        """


    @staticmethod
    def delete(paras: dict = {}) -> str:
        """
        [ 删除用例 ]
        
        ---
        可选参数:
            tid           { int } : 所属任务编号
            cid           { int } : 用例编号
            type          { str } : 类型
            state         { str } : 状态
            created_date  { str } : 创建日期, 例如 2022-02-02

        """
        # 解析搜索条件
        where_string = parse_case_search_criteria(paras)

        return f"""

        DELETE FROM { KEYWORD.DataBase.TableName.Cases }

        {where_string}

        """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 用例接口 ]

"""
import math

from flask import Response
from flask import Blueprint

from ...db import DB
from ...db import SQL
from ...api.server_request_handler import URL_PREFIX
from ...api.server_request_handler import ServerRequestHandler
from ....baseic.keyword import KEYWORD


case_blueprint = Blueprint('case', __name__)
""" 用例接口蓝图实例 """


@case_blueprint.route(''.join([URL_PREFIX, '/case/case_count']), methods=['GET'])
def case_count() -> Response:
    """
    [ 获取用例数量 ]

    ---
    可选参数:
        tid           { int } : 任务编号
        type          { str } : 类型
        state         { str } : 状态
        created_date  { str } : 创建日期, 例如 2022-02-02

    """
    try:
        # 解析请求参数
        paras = ServerRequestHandler.analysis_request_parameter(
            keys=['tid', 'type', 'state', 'created_date']
        )

        # 处理任务列表数据
        case_count = DB.read(SQL.Case.get_count(paras))[0][0]

        result = {
            'case_count': case_count
        }

        return ServerRequestHandler.successful(result)

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(str(error))


@case_blueprint.route(''.join([URL_PREFIX, '/case/case_list']), methods=['GET'])
def case_list() -> Response:
    """
    [ 获取用例列表 ]

    ---
    必传参数:
        page                { int } : 页数
        number              { int } : 单页数据量

    ---
    可选参数:
        tid           { int } : 所属任务编号
        cid           { int } : 用例编号
        type          { str } : 类型
        state         { str } : 状态
        created_date  { str } : 创建日期, 例如 2022-02-02

    """
    try:
        # 解析请求参数
        paras: dict = ServerRequestHandler.analysis_request_parameter(
            keys=['tid', 'cid', 'type', 'state', 'created_date', 'page', 'number'],
            must_keys=['page', 'number']
        )

        # 从数据库中获取原始数据
        base_case_list = DB.read(SQL.Case.get_case_list(paras))

        # 处理任务列表数据
        case_list = []
        for item in base_case_list:
            case_info = {}
            case_info.update({KEYWORD.Api.Case.Cid: item[0]})
            case_info.update({KEYWORD.Api.Case.Tid: item[1]})
            case_info.update({KEYWORD.Api.Case.Name: item[2]})
            case_info.update({KEYWORD.Api.Case.Type: item[3]})
            case_info.update({KEYWORD.Api.Case.Remark: item[4]})
            case_info.update({KEYWORD.Api.Case.State: item[5]})
            case_info.update({KEYWORD.Api.Case.CreatedDate: item[6]})
            case_info.update({KEYWORD.Api.Case.StartTime: item[7]})
            case_info.update({KEYWORD.Api.Case.EndTime: item[8]})
            case_info.update({KEYWORD.Api.Case.SpendTime: item[9]})
            case_info.update({KEYWORD.Api.Case.RunStep: item[10]})
            case_list.append(case_info)

        # 处理分页相关数据
        task_count = len(base_case_list)
        paging = {}
        page = paras['page']
        number = paras['number']
        max_page = math.ceil(task_count / number)
        paging.update({'current_page': page})
        paging.update({'current_count': number})
        paging.update({'max_page': max_page})
        paging.update({'max_count': task_count})
        paging.update({'next_page': page + 1 if page < max_page else page})
        paging.update({'back_page': page - 1 if page > 1 else 1})

        result = {
            'paging': paging,
            'tasks': case_list
        }

        return ServerRequestHandler.successful(result)

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(str(error))


@case_blueprint.route(''.join([URL_PREFIX, '/case/case_delete']), methods=['POST'])
def case_delete() -> Response:
    """
    [ 删除用例 ]

    ---
    可选参数:
        tid           { int } : 所属任务编号
        cid           { int } : 用例编号
        type          { str } : 类型
        state         { str } : 状态
        created_date  { str } : 创建日期, 例如 2022-02-02

    """
    try:
        # 解析请求参数
        paras = ServerRequestHandler.analysis_request_parameter(
            keys=['tid', 'cid', 'type', 'state', 'created_date']
        )

        # 从数据库中获取原始数据
        base_case_id_list = DB.lock_read(SQL.Case.get_id_list(paras))

        case_id_list = []
        case_id_count = 0
        del_message = '删除失败, 无法查询到用例 !'
        if len(base_case_id_list) > 0:
            for i in base_case_id_list:
                case_id_list.append(i[0])
            case_id_count = len(case_id_list)

            for cid in case_id_list:
                DB.lock_write(SQL.Case.delete({KEYWORD.Api.Case.Cid : cid}))
                DB.commit()
                del_message = f'成功删除 [{case_id_count}] 条数据 !'

        result = {
            'del_message': del_message,
            'del_count': case_id_count
        }

        return ServerRequestHandler.successful(result)

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(str(error))



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::运行服务器
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from ..awaken_info import ProjectInfo
from ..baseic.const import CONST
from ..baseic.common import window_template_output
from ..core.perform.perform_server import PerformServer


def instruction_running_server(argv: list):
    """
    [ 命令行指令::运行服务器 ]

    ---
    参数:
        argv { list } : 参数列表

    """
    if not CONST.Path.FilePath.Init.exists():
        window_template_output(
            f'Awaken2 - {ProjectInfo.Version}', 
            [
                '- 运行服务器 -',
                None,
                '当前运行路径不是工程根目录;',
                '请在工程根目录执行该命令.'
            ]
        )
        exit(0)

    port: int = 3700
    if len(argv) >= 1:
        try:
            port = int(argv[0])
        except ValueError:
            window_template_output(
                f'Awaken2 - {ProjectInfo.Version}', 
                [
                    '- 运行服务器 -',
                    None,
                    '运行服务器指令的端口参数必须为数字 !',
                ]
            )
            exit(0)

    window_template_output(
        f'Awaken2 - {ProjectInfo.Version}', 
        [
            '- 运行服务器 -',
            None,
            f'服务器正在运行 >> 端口号 :: [ {port} ]',
        ]
    )
    perform_server = PerformServer()
    perform_server.running(port)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : Awaken日志
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import time
import logging

from .const import CONST
from .config import CONFIG
from .decorator import singleton_pattern
from .error import AwakenLogCreationError
from .error import AwakenLogOutputError


class AwakenLog: ...


_LEVEL_FILE_HANDLES = [
    CONST.Common.LogLevel.Debug,
    CONST.Common.LogLevel.Info,
    CONST.Common.LogLevel.Error,
    CONST.Common.LogLevel.Warning,
    CONST.Common.LogLevel.Critical,
]
""" 文件流处理器的日志级别 """

_LOG_OUTPUT_STRUCTURE = logging.Formatter('[%(asctime)s] [%(levelname)7s] %(message)s')
""" 日志输出格式对象 """


@singleton_pattern
class AwakenLog:
    """ 
    [ 日志模块 ]

    ---
    描述:
        NULL
    
    """

    _logger: logging.Logger = logging.getLogger()
    """ 日志记录器 """


    def __init__(self) -> None:
        """ 
        [ 日志模块 :: 实例化 ]
        
        ---
        异常:
            AwakenLogCreationError : 创建日志实例异常。
        
        """
        try:
            # 从配置文件中获取日志等级
            self._logger.setLevel(CONFIG.log_level_floor.upper())
            self._logger.handlers.clear()

            # 创建以当天日志存放目录
            current_date = time.strftime('%Y-%m-%d', time.localtime())
            current_date_dir_path = CONST.Path.DirPath.Logs.joinpath(current_date)
            if not current_date_dir_path.exists():
                current_date_dir_path.mkdir(parents=True, exist_ok=True)

            # 创建文件流处理器
            for level in _LEVEL_FILE_HANDLES:
                file_handle = logging.FileHandler(current_date_dir_path.joinpath(f'{ level.upper() }.log'))
                file_handle.setFormatter(_LOG_OUTPUT_STRUCTURE)
                file_handle.setLevel(level.upper())
                self._logger.addHandler(file_handle)

            # 创建默认流处理器
            base_handle = logging.StreamHandler()
            base_handle.setFormatter(_LOG_OUTPUT_STRUCTURE)
            base_handle.setLevel(CONFIG.log_level_default.upper())
            self._logger.addHandler(base_handle)

        except BaseException as error:
            raise AwakenLogCreationError(error)


    def out_map(self, level: str, message: str):
        """ 
        [ 映射日志 ]

        ---
        描述:
            根据传递的 level 映射对应等级日志打印。

        ---
        参数:
            level   { str } : 日志等级
            message { str } : 输出的日志信息

        ---
        示例:
            >>> # 映射 error 日志, 等价于 Log.error('message')
            >>> Log = AwakenLog()
            >>> Log.out_map(level='error', message='message')

        ---
        异常:
            AwakenLogOutputError : 日志输出异常

        """
        try:
            type(self).__dict__[level](self, message)

        except KeyError as error:
            raise AwakenLogOutputError(CONST.Error.AWAKEN_LOG_OUTPUT_ERROR.replace('#LEVEL#', str(error)))


    def debug(self, message: str):
        """
        [ 记录 DEBUG 日志 ]

        ---
        参数:
            message (str) : 输出的日志信息

        """
        self._logger.debug(message)


    def info(self, message: str):
        """
        [ 记录 INFO 日志 ]

        ---
        参数:
            message {str} : 输出的日志信息

        """
        self._logger.info(message)


    def warning(self, message: str):
        """
        [ 记录 WARNING 日志 ]

        ---
        参数:
            message {str} : 输出的日志信息

        """
        self._logger.warning(message)


    def error(self, message: str):
        """
        [ 记录 ERROR 日志 ]

        ---
        参数:
            message {str} : 输出的日志信息

        """

        self._logger.error(message)
        

    def critical(self, message: str):
        """
        [ 记录 CRITICAL 日志 ]

        ---
        参数:
            message {str} : 输出的日志信息

        """
        self._logger.critical(message)


LOG: AwakenLog = AwakenLog()
""" 日志模块实例 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 全局静态配置器
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import os
import yaml

from .const import CONST
from .keyword import KEYWORD
from .decorator import singleton_pattern


class AwakenConfig: ...


@singleton_pattern
class AwakenConfig(object):
    """
    [ 全局静态配置器 ]

    ---
    描述:
        NULL

    """

    _config_cache_data: dict
    """ 配置缓存数据 """
    _debug: bool
    """ 调试模式 """
    _engine_queue_max_count: int
    """ 驱动队列阈值 """
    _task_queue_max_count: int
    """ 任务队列阈值 """
    _cpu_property_time: int
    """ CPU资源检查时间 """
    _cpu_property_ceiling: int
    """ CPU资源检查阈值 """
    _log_level_floor: str
    """ 日志等级下限 """
    _log_level_default: str
    """ 日志等级默认 """


    def __init__(self):
        """ 
        [ 全局静态配置器 :: 实例化 ] 

        ---
        描述:
            NULL

        """
        self.read_configuration_file()


    @property
    def debug(self):
        """
        [ 获取调试模式 ]

        ---
        描述:
            NULL

        """
        return self._debug


    @property
    def engine_queue_max_count(self):
        """
        [ 获取驱动队列阈值 ]

        ---
        描述:
            NULL

        """
        return self._engine_queue_max_count


    @property
    def task_queue_max_count(self):
        """
        [ 获取任务队列阈值 ]

        ---
        描述:
            NULL

        """
        return self._task_queue_max_count


    @property
    def cpu_property_time(self):
        """
        [ 获取CPU资源检查时间 ]

        ---
        描述:
            NULL

        """
        return self._cpu_property_time


    @property
    def cpu_property_ceiling(self):
        """
        [ 获取CPU资源检查阈值 ]

        ---
        描述:
            NULL

        """
        return self._cpu_property_ceiling


    @property
    def log_level_floor(self):
        """
        [ 获取日志等级下限 ]

        ---
        描述:
            NULL

        """
        return self._log_level_floor


    @property
    def log_level_default(self):
        """
        [ 获取日志等级默认 ]

        ---
        描述:
            NULL

        """
        return self._log_level_default


    def read_configuration_file(self):
        """
        [ 读取配置文件 ]

        ---
        描述:
            NULL

        """
        self._config_cache_data = {}

        # 配置文件如果为空则创建写入缺省配置数据
        if os.path.getsize(CONST.Path.FilePath.Config) == 0:
            self._write_default_data()

        with open(file=CONST.Path.FilePath.Config, mode='r', encoding='UTF-8') as file:
            self._config_cache_data      = yaml.load(file, Loader=yaml.FullLoader)
            self._debug                  = self._config_cache_data[KEYWORD.Common.Config.Debug]
            self._engine_queue_max_count = self._config_cache_data[KEYWORD.Common.Config.EngineQueueMaxCount]
            self._task_queue_max_count   = self._config_cache_data[KEYWORD.Common.Config.TaskQueueMaxCount]
            self._cpu_property_time      = self._config_cache_data[KEYWORD.Common.Config.CpuPropertyTime]
            self._cpu_property_ceiling   = self._config_cache_data[KEYWORD.Common.Config.CpuPropertyCeiling]
            self._log_level_floor        = self._config_cache_data[KEYWORD.Common.Config.LogLevelFloor]
            self._log_level_default      = self._config_cache_data[KEYWORD.Common.Config.LogLevelDefault]


    def _write_default_data(self):
        """
        [ 写入缺省数据 ]

        """
        self._config_cache_data[KEYWORD.Common.Config.Debug]               = True
        self._config_cache_data[KEYWORD.Common.Config.EngineQueueMaxCount] = 3
        self._config_cache_data[KEYWORD.Common.Config.TaskQueueMaxCount]   = 50
        self._config_cache_data[KEYWORD.Common.Config.CpuPropertyTime]     = 2
        self._config_cache_data[KEYWORD.Common.Config.CpuPropertyCeiling]  = 30
        self._config_cache_data[KEYWORD.Common.Config.LogLevelFloor]       = CONST.Common.LogLevel.Debug
        self._config_cache_data[KEYWORD.Common.Config.LogLevelDefault]     = CONST.Common.LogLevel.Info
        self.dump()


    def dump(self):
        """
        [ 存储配置 ]

        """
        with open(file=CONST.Path.FilePath.Config, mode='w', encoding='UTF-8') as file:
            yaml.dump(self._config_cache_data, file, allow_unicode=True)


    def set_debug(self, state: bool):
        """
        [ 设置调试模式 ]
        
        ---
        参数:
            state { bool } : 调试模式

        """
        self._debug = state
        self._config_cache_data[KEYWORD.Common.Config.Debug] = state


    def set_engine_queue_max_count(self, number: int):
        """
        [ 设置驱动队列阈值 ]
        
        ---
        参数:
            number { int } : 数量

        """
        self._engine_queue_max_count = number
        self._config_cache_data[KEYWORD.Common.Config.EngineQueueMaxCount] = number


    def set_task_queue_max_count(self, number: int):
        """
        [ 设置任务队列阈值 ]
        
        ---
        参数:
            number { int } : 数量

        """
        self._task_queue_max_count = number
        self._config_cache_data[KEYWORD.Common.Config.TaskQueueMaxCount] = number


    def set_cpu_property_time(self, property_time: int):
        """
        [ 设置CPU资源检查时间 ]
        
        ---
        参数:
            property_time { int } : 检测时间

        """
        self._cpu_property_time = property_time
        self._config_cache_data[KEYWORD.Common.Config.CpuPropertyTime] = property_time


    def set_cpu_property_ceiling(self, capacity_size: int):
        """
        [ 设置CPU资源检查阈值 ]
        
        ---
        参数:
            capacity_size { int } : 检查阈值

        """
        self._cpu_property_ceiling = capacity_size
        self._config_cache_data[KEYWORD.Common.Config.CpuPropertyCeiling] = capacity_size


    def set_log_level_floor(self, log_level: str):
        """
        [ 设置日志等级下限 ]

        ---
        参数:
            log_level { str } : 日志等级

        """
        try:
            self._log_level_floor = CONST.Common.LogLevel.__dict__[log_level]

        except KeyError:
            self._log_level_floor = CONST.Common.LogLevel.Debug

        self._config_cache_data[KEYWORD.Common.Config.Debug] = self._log_level_floor


    def set_log_level_default(self, log_level: str):
        """
        [ 设置日志等级默认 ]

        ---
        参数:
            log_level { str } : 日志等级

        """
        try:
            self.log_level_default = CONST.Common.LogLevel.__dict__[log_level]

        except KeyError:
            self.log_level_default = CONST.Common.LogLevel.Info

        self._config_cache_data[KEYWORD.Common.Config.Debug] = self.log_level_default


CONFIG: AwakenConfig = AwakenConfig()
""" 静态配置实例 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::版本
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import os
import re
import subprocess

from ..awaken_info import ProjectInfo
from ..baseic.common import window_template_output


def instruction_version():
    window_template_output(
        f'Awaken2 - {ProjectInfo.Version}', 
        [
            '- 检查版本 -',
            None,
            '正在查询 PYPI 服务器...',
        ]
    )

    dos_result = subprocess.Popen(f'pip install {ProjectInfo.Name}==', stderr=subprocess.PIPE)
    _, err = dos_result.communicate()
    newest_version = re.findall(r'from versions: (.*)\)', str(err))[0].split(',')[-1].strip()
    window_template_output(
        f'Awaken2 - {ProjectInfo.Version}', 
        [
            '- 检查版本 -',
            None,
            f'当前版本: { ProjectInfo.Version }',
            f'最新版本: { newest_version }',
        ]
    )
    exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : API字段关键字
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""


class Task: ...
class Case: ...


class Task:
    """
    [ API字段关键字索引 :: 任务 ]

    ---
    描述: 
        任务相关的字段。

    """

    Tid = 'tid'
    """ 编号 """

    Name = 'name'
    """ 名称 """

    Type = 'type'
    """ 类型 """

    Remark = 'remark'
    """ 备注 """

    CreatedDate = 'created_date'
    """ 创建日期 """

    State = 'state'
    """ 状态 """

    StartTime = 'start_time'
    """ 开始时间 """

    EndTime = 'end_time'
    """ 结束时间 """

    SpendTime = 'spend_time'
    """ 消耗时间 """

    RecordCasesAll = 'record_cases_all'
    """ 统计所有用例数 """

    RecordCasesSuccess = 'record_cases_success'
    """ 统计成功用例数 """

    RecordCasesUnSuccess = 'record_cases_unsuccess'
    """ 统计失败用例数 """


class Case:
    """
    [ API字段关键字索引 :: 用例 ]

    ---
    描述: 
        用例相关的字段。
        
    """

    Cid = 'cid'
    """ 用例编号 """

    Tid = Task.Tid
    """ 所属任务编号 """

    Name = Task.Name
    """ 名称 """

    Type = Task.Type
    """ 类型 """

    Remark = Task.Remark
    """ 备注 """

    CreatedDate = Task.CreatedDate
    """ 创建日期 """

    State = Task.State
    """ 状态 """

    StartTime = Task.StartTime
    """ 开始时间 """

    EndTime = Task.EndTime
    """ 结束时间 """

    SpendTime = Task.SpendTime
    """ 消耗时间 """

    RunStep = 'run_step'
    """ 执行步骤 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 服务端相关 ]

"""



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 数据接口 ]

"""
from flask import Response
from flask import Blueprint

from ...db import DB
from ...db import SQL
from ...api.server_request_handler import URL_PREFIX
from ...api.server_request_handler import ServerRequestHandler


data_blueprint = Blueprint('data', __name__)
""" 数据接口蓝图实例 """


@data_blueprint.route(''.join([URL_PREFIX, '/data/summarize']), methods=['GET'])
def summarize() -> Response:
    """
    [ 获取执行信息概述 ]

    """
    try:
        # 获取最新执行数据
        new_exec_data = {}
        # 获取日期
        date = DB.read(SQL.TASK.get_date_list())[0][0]
        new_exec_data.update({'date': date})
        # 获取任务数量
        new_exec_data.update({'task_count': DB.read(SQL.TASK.get_count_from_date(date)[0][0])})
        # 获取用例数量
        new_exec_data.update({'case_count': DB.read(SQL.CASE.get_count_from_date(date))[0][0]})
        # 获取异常用例数量
        new_exec_data.update({'fail_case_count': DB.read(SQL.CASE.get_count_fail_from_data(date))[0][0]})
        # 获取消耗时间
        new_exec_data.update({'spend_time': round((DB.read(SQL.TASK.get_spend_time_from_data(date))[0][0] / 60), 2)})

        # 获取历史执行数据
        history_exec_data = {}
        # 获取任务数量
        history_exec_data.update({'task_count': DB.read(SQL.TASK.get_count_all())[0][0]})
        # 获取用例数量
        history_exec_data.update({'case_count': DB.read(SQL.CASE.get_count_all())[0][0]})
        # 获取异常用例数量
        history_exec_data.update({'fail_case_count': DB.read(SQL.CASE.get_count_fail())[0][0]})
        # 获取消耗时间
        history_exec_data.update({'spend_time': round((DB.read(SQL.TASK.get_spend_time_all())[0][0] / 60), 2)})
        # 获取异常任务数量
        history_exec_data.update({'fail_task_count': DB.read(SQL.TASK.get_count_fail())[0][0]})

        return ServerRequestHandler.successful({
            'new_exec': new_exec_data, 
            'history_exec': history_exec_data
        })

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(error)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 路径常量
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from pathlib import Path


class DirPath: ...
class FilePath: ...


CWD = Path().cwd()
""" Path对象 :: 运行时路径 """


class DirPath:
    """
    [ 路径常量索引 :: 目录路径对象 ]

    ---
    描述: 
        NULL

    """

    Data = CWD.joinpath('Data')
    """ 数据目录 """

    Logs = Data.joinpath('Logs')
    """ 日志目录 """

    BaseCode = Data.joinpath('BaseCode')
    """ 底层编码目录 """


class FilePath:
    """
    [ 路径常量索引 :: 文件路径对象 ]

    ---
    描述: 
        NULL
        
    """

    Init = CWD.joinpath('AwakenEngineering.ini')

    Config = DirPath.Data.joinpath('AwakenConfig.yaml')
    """ 项目配置文件 """

    Database = DirPath.Data.joinpath('AwakenDatabase.db')
    """ 项目数据库文件 """





# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 状态常量
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""


class Result: ...
class Switch: ...
class RunProcess: ...


class Result:
    """
    [ 状态常量索引 :: 结果 ]

    ---
    描述: 
        NULL

    """

    Null = 'Null'
    """ NULL """

    Success = '成功'
    """ 成功 """

    Unsuccess = '失败'
    """ 失败 """

    Error = '异常'
    """ 异常 """


class Switch:
    """
    [ 状态常量索引 :: 开关 ]

    ---
    描述: 
        NULL

    """

    On = '开'
    """ 开 """

    Off = '关'
    """ 关 """


class RunProcess:
    """
    [ 状态常量索引 :: 运行过程 ]

    ---
    描述: 
        NULL
        
    """
    
    Start = '开始'
    """ 开始 """

    End = '结束'
    """ 结束 """

    Suspend = '暂停'
    """ 暂停 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : Baseic.Common
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import os
import hashlib
import subprocess

from pathlib import Path

from .const import CONST
from ..awaken_info import ProjectInfo


_ENVIRONMENT_VARIABLE = 'USERPROFILE'
_PLAYWRIGHT_PATH = Path(os.environ[_ENVIRONMENT_VARIABLE]).joinpath('AppData\Local\ms-playwright')

TASK_TEMPLATE_WEB = \
"""
## 自述
TaskName = ''
TaskDocs = ''
TaskType = 'WEB'


## 用例
CASE :: TestName :: Docs
    Goto >> 'http://www.baidu.com/'
    ?? True
"""

TASK_TEMPLATE_MAP = {
    CONST.Type.Task.Web : TASK_TEMPLATE_WEB
}


def check_browser_driver_is_installed():
    """
    [ 检查浏览器驱动是否安装 ]

    ---
    描述:
        NULL

    """
    browser_list = ['chromium', 'firefox', 'webkit']
    local_browser_list = []

    # 检查本地 playwright 环境是否安装浏览器驱动
    if _ENVIRONMENT_VARIABLE in os.environ:
        for dir_path in _PLAYWRIGHT_PATH.glob('*'):
            local_browser_list.append(dir_path.stem.split('-')[0])

        # 如果浏览器驱动不存在本地则下载
        for browser_name in browser_list:
            if browser_name not in local_browser_list:
                dos_result = subprocess.Popen(['pip', 'install', f'{ browser_name }'], stderr=subprocess.PIPE)
                dos_result.communicate()
                window_template_output(
                    f'Awaken2 - {ProjectInfo.Version}', 
                    [
                        '- 检查本地浏览器驱动 -',
                        None,
                        f'正在下载浏览器驱动 >> {browser_name}'
                    ]
                )
                os.system(f'playwright install { browser_name }')


def create_program_runtime_dependency_files():
    """
    [ 创建程序运行时的依赖文件 ]

    ---
    描述:
        NULL
    
    """
    for path in [
        CONST.Path.DirPath.Data,
        CONST.Path.DirPath.Logs,
        CONST.Path.DirPath.BaseCode,
    ]:
        path.mkdir(parents=True, exist_ok=True)

    for path in [
        CONST.Path.FilePath.Config,
        CONST.Path.FilePath.Database,
    ]:
        path.touch(mode=0o777, exist_ok=True)


def create_project(name: str):
    """ 
    [ 创建项目 ]

    ---
    参数:
        name { str } : 项目名称

    ---
    返回:
        bool : 是否创建成功凭证
    
    """
    project_path = CONST.Path.CWD.joinpath(name)

    if not project_path.exists():
        project_path.mkdir(parents=True, exist_ok=True)
        return True
    else:
        return False


def create_task(task_name: str, task_type: str):
    """ 
    [ 创建任务 ]

    ---
    参数:
        name { str } : 任务名称

    ---
    返回:
        bool : 是否创建成功凭证
    
    """
    task_name = f'{task_name}.awaken-{task_type}'
    task_path = CONST.Path.CWD.joinpath(task_name)

    if not task_path.exists():
        task_path.touch(mode=0o777)
        task_path.write_bytes(bytes(TASK_TEMPLATE_MAP[task_type.upper()], encoding='UTF8'))
        return True
    else:
        return False


def md5_encrypt(string: str):
    """
    [ MD5加密 ]
    
    ---
    描述:
        NULL

    ---
    参数:
        string { str } : 需要加密的字符串。

    """
    file_md5hash = hashlib.md5(string.encode('UTF-8'))
    return file_md5hash.hexdigest()


def window_template_output(name: str, message: list):
    os.system('cls')
    max_limit = 50

    def calculate_offset(value):
        offset = 0
        for s in value:
            if u'\u4e00' <= s <= u'\u9fff':
                offset += 1
        return offset

    print('┏' + '━'*max_limit + '┓')
    print('┃ ' + name + ' '*((max_limit - len(name) - 2) - calculate_offset(name)) + ' ┃')
    print('┣' + '━'*max_limit + '┫')

    print('┃' + ' '*max_limit + '┃')
    for mage in message:
        if not mage:
            print('┃' + ' '*max_limit + '┃')
        else:
            print('┃ ' + mage + ' '*((max_limit - len(mage) - 2) - calculate_offset(mage)) + ' ┃')
    print('┃' + ' '*max_limit + '┃')

    print('┗' + '━'*max_limit + '┛')



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 关键字索引
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from . import keyword_api
from . import keyword_common
from . import keyword_script
from . import keyword_database


class AwakenKeyword: ...


class AwakenKeyword:
    """
    [ 关键字索引 ]

    ---
    描述:
        NULL

    """


    @property
    def Api(self) -> keyword_api:
        """
        [ API字段关键字 ]

        ---
        描述:
            NULL

        """
        return keyword_api


    @property
    def Common(self) -> keyword_common:
        """
        [ 通用字段关键字 ]

        ---
        描述:
            NULL

        """
        return keyword_common


    @property
    def Script(self) -> keyword_script:
        """
        [ 脚本字段关键字 ]

        ---
        描述:
            NULL

        """
        return keyword_script


    @property
    def DataBase(self) -> keyword_database:
        """
        [ 数据库关键字 ]

        ---
        描述:
            NULL

        """
        return keyword_database


KEYWORD: AwakenKeyword = AwakenKeyword()
""" 关键字索引实例 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 解释器常量
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""


class StatementIdent: ...
class GrammarSymbol: ...
class CodeLineType: ...
class CodeLineScopet: ...


KEYWORD_BASECODE_FILE_SUFFIX = 'basecode'
"""
[ 关键字 :: 编码文件后缀 ]
描述: 底层编码文件的命名拼接字段。
"""

KEYWORD_IDENT_SCOPE_UNIVERSE = 'UNIVERSE'
""" 
[ 关键字 :: 公共作用域标识  ]
描述: 底层编码行的作用域标识, 表示作用于全域。
"""

SYMBOL_BASECODE_FILE_SEPARATOR = '-'
""" 
[ 符号 :: 编码文件间隔符  ]
描述: 底层编码文件的命名拼接间隔符号。
"""


class StatementIdent:
    """
    [ 解释器常量索引 :: 声明标识 ]

    ---
    描述: 
        底层编码行的声明事件参数标识。

    """

    Case = 'CASE'
    """ 声明用例标识 """

    Deco = 'DECO'
    """ 声明装饰器标识 """


class GrammarSymbol:
    """
    [ 解释器常量索引 :: 语法符号 ]

    ---
    描述: 
        Awaken脚本语法中的语法符号。

    """

    Annotation = '##'
    """ 注释符 """

    Give = '='
    """ 赋值符 """

    Statement = '::'
    """ 声明符 """

    Run = '>>'
    """ 执行符 """

    Assert = '??'
    """ 断言符 """

    Call = '@'
    """ 方法调用符 """

    Quote = '$'
    """ 资源引用符 """

    VariablePath = '::'
    """ 字典路径应用符 """

    ScopePrefix = '?>'
    """ 作用域前缀 """


class CodeLineType:
    """
    [ 解释器常量索引 :: 编码行类型 ]

    ---
    描述: 
        NULL

    """

    Give = 'GIVE'
    """ 赋值类型 """

    RGive = 'RGIVE'
    """ 执行赋值类型 """

    SCase = 'SCASE'
    """ 声明用例类型 """

    SDecorator = 'SDECORATOR'
    """ 声明装饰器类型 """

    Run = 'RUN'
    """ 执行类型 """

    Assert = 'ASSERT'
    """ 断言类型 """


class CodeLineScopet:
    """
    [ 解释器常量索引 :: 编码行作用域 ]

    ---
    描述:
        NULL

    """

    Global = 'GLOBAL'
    """ 全局 """

    Local = 'LOCAL'
    """ 局部 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行工具
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    该模块适用于辅助 Awaken 进行命令行调用。

"""
import sys

from .awaken_info import ProjectInfo
from .baseic.common import window_template_output

from .external.order import instruction_order
from .external.init_engineering import instruction_init_engineering
from .external.create_project import instruction_create_project
from .external.create_task import instruction_create_task
from .external.running_task import instruction_running_task
from .external.running_server import instruction_running_server
from .external.help import instruction_help
from .external.version import instruction_version
from .external.update import instruction_update


# ----------------------------------------------------------------------------
# 命令行工具入口
# ----------------------------------------------------------------------------
def main():
    command = 'order'
    
    if len(sys.argv) >= 2:
        command = sys.argv[1]
        argv    = sys.argv[2:]

    # ------------------------------
    if command in ['order', '-order', '-o']:
        instruction_order()

    # ------------------------------
    elif command in ['init', '-init', '-i']:
        instruction_init_engineering()

    # ------------------------------
    elif command in ['make', '-make', '-m']:
        instruction_create_project(argv)

    # ------------------------------
    elif command in ['task', '-task', '-t']:
        instruction_create_task(argv)

    # ------------------------------
    elif command in ['run', '-run', '-r']:
        instruction_running_task(argv)

    # ------------------------------
    elif command in ['server', '-server', '-s']:
        instruction_running_server(argv)

    # ------------------------------
    elif command in ['help', '-help', '-h']:
        instruction_help()

    # ------------------------------
    elif command in ['version', '-version', '-v']:
        instruction_version()

    # ------------------------------
    elif command in ['update', '-update', '-u']:
        instruction_update()

    else:
        window_template_output(
            f'Awaken2 - {ProjectInfo.Version}',
            [
                f'没有这样的指令:: [{command}] !'
            ]
        )
        exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 脚本全局方法
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import yaml
import time
from pathlib import Path

from ..baseic.log import LOG
from ..baseic.const import CONST
from ..baseic.decorator import singleton_pattern


class GlobalMethod: ...


@singleton_pattern
class GlobalMethod:
    """
    [ 脚本全局方法 ]
    
    """

    def test(self):
        """
        [ 测试 ]

        """
        print('成功!')


    def sleep(self, wait: int):
        """
        [ 睡眠 ]

        ---
        描述:
            NULL

        """
        time.sleep(wait)


    def logout(self, message: str, level=CONST.Common.LogLevel.Info) -> str:
        """
        [ 日志输出 ]

        ---
        描述:
            输出日志到控制台。
        
        ---
        参数:
            message { str } : 输出的日志信息。
            level   { str } : 日志等级。
        
        """
        LOG.out_map(level, message)
        return message


    def read_yaml(self, yaml_path_str: str):
        """
        [ 读取YAML ]

        ---
        描述:
            NULL

        """
        yaml_path_str = Path(yaml_path_str)
        with open(file=yaml_path_str, mode='r', encoding='UTF-8') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        return data



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 数据库模块通用函数 ]

"""
from ...baseic.keyword import KEYWORD


def parse_task_search_criteria(paras: dict):
    """
    [ 解析任务搜索条件 ]
    
    """
    where_list = []
    paras_keys = paras.keys()
    if KEYWORD.Api.Task.Tid in paras_keys:
        where_list.append(f'{KEYWORD.Api.Task.Tid} = {paras[KEYWORD.Api.Task.Tid]}')
    if KEYWORD.Api.Task.Type in paras_keys:
        where_list.append(f'{KEYWORD.Api.Task.Type} = "{paras[KEYWORD.Api.Task.Type]}"')
    if KEYWORD.Api.Task.Name in paras_keys:
        where_list.append(f'{KEYWORD.Api.Task.Name} = "{paras[KEYWORD.Api.Task.Name]}"')
    if KEYWORD.Api.Task.State in paras_keys:
        where_list.append(f'{KEYWORD.Api.Task.State} = "{paras[KEYWORD.Api.Task.State]}"')
    if KEYWORD.Api.Task.CreatedDate in paras_keys:
        where_list.append(f'{KEYWORD.Api.Task.CreatedDate} = "{paras[KEYWORD.Api.Task.CreatedDate]}"')
    if 'record_cases_state' in paras_keys:
        if paras['record_cases_state'] == 0:
            where_list.append(f'{KEYWORD.Api.Task.RecordCasesUnSuccess} = 0')
        elif paras['record_cases_state'] == 1:
            where_list.append(f'{KEYWORD.Api.Task.RecordCasesUnSuccess} != 0')

    return splicing_sql_where_string(where_list)


def parse_case_search_criteria(paras: dict):
    """
    [ 解析任务搜索条件 ]
    
    """
    where_list = []
    paras_keys = paras.keys()
    if KEYWORD.Api.Case.Cid in paras_keys:
        where_list.append(f'{KEYWORD.Api.Case.Cid} = {paras[KEYWORD.Api.Case.Cid]}')
    if KEYWORD.Api.Case.Tid in paras_keys:
        where_list.append(f'{KEYWORD.Api.Case.Tid} = {paras[KEYWORD.Api.Case.Tid]}')
    if KEYWORD.Api.Case.Type in paras_keys:
        where_list.append(f'{KEYWORD.Api.Case.Type} = "{paras[KEYWORD.Api.Case.Type]}"')
    if KEYWORD.Api.Case.Name in paras_keys:
        where_list.append(f'{KEYWORD.Api.Case.Name} = "{paras[KEYWORD.Api.Case.Name]}"')
    if KEYWORD.Api.Case.State in paras_keys:
        where_list.append(f'{KEYWORD.Api.Case.State} = "{paras[KEYWORD.Api.Case.State]}"')
    if KEYWORD.Api.Case.CreatedDate in paras_keys:
        where_list.append(f'{KEYWORD.Api.Case.CreatedDate} = "{paras[KEYWORD.Api.Case.CreatedDate]}"')

    return splicing_sql_where_string(where_list)


def get_desc_limit(page: int = 1, number: int = 20) -> str:
    """
    [ 获取数据返回量区间限制 ]

    ---
    描述:
        默认是获取 1 页, 20 条数据。
    
    ---
    参数:
        page { int } : 页数。
        number { int } : 单页数据量。

    ---
    返回:
        Response : 拼接完成的 SQL LIMIT 语句参数。

    """
    if not isinstance(page, int):
        page = int(page)
    if not isinstance(number, int):
        number = int(number)
    
    # 计算偏移量
    limit_begin = 0
    if page > 1:
        limit_begin = (page - 1) * number

    return f'{ limit_begin }, { number }'


def splicing_sql_where_string(where_list: list) -> str:
    """
    [ 拼接 SQL 搜索条件 ]

    """
    where_key_str = ''
    number = 0
    end_number = len(where_list)
    if end_number >= 1:
        where_key_str += 'WHERE '
        for where in where_list:
            if number > 0:
                where_key_str += ' AND '
            where_key_str += where
            number += 1
    return where_key_str



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ make测试接口 ]

"""
from flask import jsonify
from flask import Blueprint

from ...db import DB
from ...db import SQL
from ...api.server_request_handler import URL_PREFIX
from ...api.server_request_handler import ServerRequestHandler


BLUEPRINT_TEST = Blueprint('test', __name__)
""" [ 测试接口蓝图 ] """

BASE_USERS = {
    'admin': {
        'userId': 1, 
        'username': 'admin', 
        'realName': 'Rains Admin', 
        'avatar': 'http://mms0.baidu.com/it/u=2163504278, 3640533387&fm=253&app=138&f=JPEG&fmt=auto&q=75?w=500&h=707', 
        'desc': 'manager', 
        'password': 'admin', 
        'token': 'fakeToken1', 
        'homePath': '/dashboard/workbench', 
        'roles': [
            {
                'roleName': 'Super Admin', 
                'value': 'super'
            }, 
        ], 
    }
}
""" [ 临时用户数据 ] """


@BLUEPRINT_TEST.route(f'{URL_PREFIX}/login', methods=['POST'])
def get() -> jsonify:
    try:
        paras: dict = ServerRequestHandler.analysis_request_parameter(keys=['username', 'password'])

        if paras['username'] in BASE_USERS.keys():
            user = BASE_USERS[paras['username']]
            if user['password'] == paras['password']:
                result = {
                    'roles': user['roles'], 
                    'userId': user['userId'], 
                    'username': user['username'], 
                    'token': user['token'], 
                    'realName': user['realName'], 
                    'desc': user['desc'], 
                }
                return jsonify({
                    'code': 0, 
                    'result': result, 
                    'message': 'ok', 
                    'type': 'success'
                })
        else:
            return jsonify({
                'code': -1, 
                'result': None, 
                'message': 'Incorrect account or password！', 
                'type': 'error'
            })

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(f'{ error }')


@BLUEPRINT_TEST.route(f'{URL_PREFIX}/getUserInfo', methods=['GET'])
def get_user_info() -> jsonify:
    try:
        headers: dict = ServerRequestHandler.analysis_request_headers(keys=['Authorization'], must_keys=['Authorization'])
        token = headers['Authorization']

        if token:
            for _, v in BASE_USERS.items():
                if token == v['token']:
                    return jsonify({
                        'code': 0, 
                        'result': v, 
                        'message': 'ok', 
                        'type': 'success'
                    })
        else:
            return jsonify({
                'code': -1, 
                'result': None, 
                'message': 'Invalid token!', 
                'type': 'error'
            })

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(f'{ error }')


@BLUEPRINT_TEST.route(f'{URL_PREFIX}/logout', methods=['GET'])
def logout() -> jsonify:
    try:
        headers: dict = ServerRequestHandler.analysis_request_headers(keys=['Authorization'], must_keys=['Authorization'])
        token = headers['Authorization']

        if token:
            for k, v in BASE_USERS.items():
                if token == v['token']:
                    return jsonify({
                        'code': 0, 
                        'result': v, 
                        'message': 'Token has been destroyed', 
                        'type': 'success'
                    })
        else:
            return jsonify({
                'code': -1, 
                'result': None, 
                'message': 'Invalid token!', 
                'type': 'error'
            })

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(f'{ error }')


@BLUEPRINT_TEST.route(f'{URL_PREFIX}/table/getDemoList', methods=['GET'])
def table() -> jsonify:
    try:
        paras: dict = ServerRequestHandler.analysis_request_parameter(keys=['page', 'pageSize'])
        page = paras['page']
        page_size = paras['pageSize']

        headers: dict = ServerRequestHandler.analysis_request_headers(keys=['Authorization'], must_keys=['Authorization'])
        token = headers['Authorization']

        # 获取服务器数据
        db_r = DB.read(SQL.TASK.get_all_item(int(page), int(page_size)))
        task_all_count = DB.read(SQL.TASK.get_count_all())[0][0]
 
        r_task = []
        for t in db_r:
            task = {}
            task.update({'tid': t[0]})
            task.update({'name': t[1]})
            task.update({'remark': t[2]})
            task.update({'created_date': str(t[3])})
            task.update({'state': t[4]})
            task.update({'start_time': t[5]})
            task.update({'end_time': t[6]})
            task.update({'spend_time_s': t[7]})
            r_task.append(task)

        if token:
            for _, v in BASE_USERS.items():
                if token == v['token']:
                    return jsonify({
                        'code': 0, 
                        'result': {
                            'items': r_task, 
                            'total': task_all_count
                        }, 
                        'message': 'ok', 
                        'type': 'success'
                    })

    except BaseException as error:
        return ServerRequestHandler.unsuccessful(f'{ error }')



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 错误常量
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""


class Interpreter: ...
class PerformPool: ...


AWAKEN_LOG_OUTPUT_ERROR = '日志等级不存在 :: [ #LEVEL# ] !'


class Interpreter:
    """
    [ 错误常量 :: 解释器 ]

    ---
    描述: 
        NULL
        
    """

    GrammarSymbolWrongful = '语法符号不合法 :: [ #SYMBOL# ] !'
    """ 语法符号不合法 """

    GrammarMethodNotCallSymbol = '请使用 "#CALL##CALL#" 调用全局方法, 在用例空间中使用 "#CALL#" 调用引擎方法 !'
    """ 语法方法未使用调用符号 """

    GrammarGlobalProhibitUsingLocalMethod = '全域命名空间中无法调用引擎方法, 请使用 "#CALL##CALL#" 调用全局方法 !'
    """ 语法全局禁止使用本地方法 """

    GrammarStatementWrongful = '声明关键字 [ #KEYWORD# ] 不是合法的关键字 !'
    """ 语法声明关键字不合法 """

    GrammarStatementNoParameters = '声明关键字 [ #KEYWORD# ] 没有携带参数 !'
    """ 语法声明关键字无参数 """

    GrammarUseCaseNoAssertion = '用例 [ #CASE_NAME# ] 没有断言语句 !'
    """ 语法用例未断言 """

    GrammarBaseCodeLineTypeError = '无法识别的编码行类型 :: [ #TYPE# ] !'
    """ 语法底层编码行类型异常 """

    GrammarMethodWrongful = 'Awaken语法异常 :: 代码行 [ #LINE# ] 引用了不合法的方法 :: [ #NAME# ] !'
    """ 语法全局方法不合法 """

    GrammarNamespaceNodeNotExist = '命名空间不存在节点 [ #NODE# ]'
    """ 语法命名空间不存在节点 """


class PerformPool:
    """
    [ 错误常量 :: 执行器 ]

    ---
    描述: 
        NULL
        
    """

    TaskQueueConsumedMessage = '任务队列消耗完毕, 正在等待任务执行 !'

    TaskRunningConsumedMessage = '任务全部执行完毕, 退出程序 !'

    ServerRunningMessage = '执行服务运行中 >> 端口号 :: [ #PORT# ] !'

    ListeningProcessStarted = '监听进程已启动 ...'



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 解释器通用方法
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import re


def converting_para_format(para: str):
    """
    [ 转换参数格式 ]

    ---
    描述:
        NULL

    """
    # 参数为空则返回
    if para == '': 
        return ''
    para = str(para)

    # 解析字符串
    if para[0] in ["'", '"'] and para[-1] in ["'", '"']:
        string_extract = re.findall(r'\'(.*)\'', para)
        if len(string_extract) < 1:
            string_extract = re.findall(r'\"(.*)\"', para)
        if len(string_extract) >= 1:
            return string_extract[0]

    else:
        # 解析BOOL值
        if para in ['true', 'True', 'TRUE']:
            return True
        if para in ['false', 'False', 'FALSE']:
            return False

        # 解析数字
        try:
            int_extract = float(para)
            if int_extract.is_integer():
                int_extract = int(int_extract)
            return int_extract

        except ValueError:
            return para



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ API驱动引擎 ]

"""
import time
import requests

from ...baseic.const import CONST
from ...baseic.log import LOG
from ...baseic.error import AwakenApiEngineError


class ApiEngine: ...


class ApiEngine(object):
    """
    [ API驱动引擎 ]

    ---
    描述:
        NULL

    """

    _headers: dict
    """ 请求头 """


    def __init__(self) -> None:
        """
        [ API驱动引擎 :: 初始化 ]

        ---
        描述:
            NULL

        """
        self._headers = {}


    def logout(self, message: str, level=CONST.Common.LogLevel.Info) -> str:
        """
        [ 日志输出 ]

        ---
        描述:
            输出日志到控制台。
        
        ---
        参数:
            message { str } : 输出的日志信息。
            level   { str } : 日志等级。
        
        """
        LOG.out_map(level, message)
        return message


    def sleep(self, wait: int):
        """
        [ 停滞等待 ]

        ---
        描述:
            NULL
        
        ---
        参数:
            wait { int } : 等待时长, 单位秒。
        
        """
        time.sleep(int(wait))


    def headers_add(self, key: str, value: str):
        """
        [ 请求头新增 ]

        ---
        描述:
            NULL

        """
        self._headers.update({key : value})


    def set_token(self, token: str):
        """
        [ 设置登录态 ]

        ---
        描述:
            NULL

        """
        self._headers.update({'Authorization' : f'Bearer {token}'})


    def get(self, url: str, data: dict = None):
        """
        [ GET 请求 ]

        ---
        描述:
            NULL

        """
        result = requests.get(url=url, params=data, headers=self._headers)
        return result.json()


    def post(self, url: str, data: dict = None):
        """
        [ POST 请求 ]

        ---
        描述:
            NULL
            
        """
        result = requests.post(url=url, data=data, headers=self._headers)
        return result.json()



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 解释器结构体
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from .common import converting_para_format
from ...baseic.const import CONST
from ...baseic.error import AwakenConvertCodelinError
from ...kit.global_method import GlobalMethod


class AwakenTask: ...
class AwakenCase: ...
class AwakenCodeLine: ...


class AwakenTask:
    """
    [ Awaken任务对象 ]

    ---
    描述:
        该对象描述自动化任务的完整执行信息。

    """

    global_method: GlobalMethod = GlobalMethod()
    global_function_map: dict
    engine_function_map: dict
    namespace: dict
    test_cases: dict
    task_type: str
    basecodes: list


    def __init__(self, task_type: str, basecodes: list) -> None:
        """
        [ Awaken任务对象 :: 初始化 ]

        """
        self.task_type = task_type
        self.basecodes = basecodes


class AwakenCase:
    """
    [ Awaken用例 ]

    ---
    描述:
        该对象描述自动化任务用例的信息。

    """

    number: int
    name: str
    docs: str
    decorator: dict
    namespace: dict
    steps: list[AwakenCodeLine]


    def __init__(self, number, name, docs='', *_) -> None:
        self.number = number
        self.name = name
        self.docs = docs
        self.decorator = {}
        self.namespace = {}
        self.steps = []


class AwakenCodeLine:
    """
    [ 代码行对象 ]

    ---
    描述:
        该对象描述自动化任务用例的信息。

    """

    number: int
    region: str
    type: str
    give_region: str
    give_name: str
    give_value: str
    funtion_region: str
    funtion_name: str
    funtion_value: str
    case_name: str
    case_docs: str
    decorator_key: str
    decorator_value: str
    assert_value: str


    def __init__(self, codeline: str) -> None:
        codeline_splint = codeline.split(' ')
        self.number = codeline_splint[0]
        self.region = codeline_splint[1]
        self.type = codeline_splint[2]

        if self.type == CONST.Interpreter.CodeLineType.Give:
            _give_splint = codeline_splint[3].split(CONST.Interpreter.GrammarSymbol.ScopePrefix)
            self.give_region = _give_splint[0]
            self.give_name = _give_splint[1]
            self.give_value = converting_para_format(codeline_splint[4])

        elif self.type == CONST.Interpreter.CodeLineType.RGive:
            _give_splint = codeline_splint[3].split(CONST.Interpreter.GrammarSymbol.ScopePrefix)
            _run_splint = codeline_splint[4].split(CONST.Interpreter.GrammarSymbol.ScopePrefix)
            self.give_region = _give_splint[0]
            self.give_name = _give_splint[1]
            self.funtion_region = _run_splint[0]
            self.funtion_name = _run_splint[1]
            self.funtion_value = []
            if len(codeline_splint) >= 6:
                for value in codeline_splint[5:]:
                    self.funtion_value.append(converting_para_format(value))

        elif self.type == CONST.Interpreter.CodeLineType.Run:
            _run_splint = codeline_splint[3].split(CONST.Interpreter.GrammarSymbol.ScopePrefix)
            self.funtion_region = _run_splint[0]
            self.funtion_name = _run_splint[1]
            self.funtion_value = []
            if len(codeline_splint) >= 5:
                for value in codeline_splint[4:]:
                    self.funtion_value.append(converting_para_format(value))

        elif self.type == CONST.Interpreter.CodeLineType.SCase:
            self.case_name = codeline_splint[3]
            self.case_docs = codeline_splint[4] if len(codeline_splint) == 5 else ''

        elif self.type == CONST.Interpreter.CodeLineType.SDecorator:
            self.decorator_key = codeline_splint[2]
            self.decorator_value = codeline_splint[3]

        elif self.type == CONST.Interpreter.CodeLineType.Assert:
            self.assert_value = codeline_splint[3:]

        else:
            raise AwakenConvertCodelinError(CONST.Error.Interpreter.GrammarBaseCodeLineTypeError.replace('#TYPE#', self.type))



# Copyright 2022. quinn.7@foxmail.com All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ API 执行器 ]

"""
import re
import time

from .api_engine import ApiEngine
from ..common import _replace_variables_in_parameters
from ...base.log import LOG
from ...base.const import CONST
from ...base.language import LANGUAGE
from ...base.error import AwakenTaskRunningError
from ...base.error import AwakenTaskCaseAssertError
from ...server.task_db_broker import TaskDbBroker
from ...core.interpreter.grammar_parser import AwakenTask

class ApiRunner: ...


class ApiRunner:
    """
    [ API 执行器 ]

    """
    _task: AwakenTask
    _engine_action_map: dict
    _task_db_broker: TaskDbBroker


    def __init__(self, task: AwakenTask) -> None:
        self._task = task
        self._engine_action_map = {}
        self._task_db_broker = TaskDbBroker(CONST.TYPE.TYPE_API)

        # 解析获取驱动关键字对应的函数映射字典
        for n, v in ApiEngine.__dict__.items():
            if n[0] != '_':
                n = ''.join([n.title() for n in n.rsplit('_')])
                self._engine_action_map.update({ n : v })


    def start(self, api_engine: ApiEngine):
        """
        [ 开始任务 ]

        """
        task_state = CONST.TEXT.TEXT_END

        # 数据库创建任务
        self._task_db_broker.task_start(
            self._task.setting['TaskName'].value,
            self._task.setting['TaskDocs'].value
        )

        # 遍历任务的用例
        for case_object in self._task.test_cases:

            # 数据库创建用例
            case_id = self._task_db_broker.created_case(
                case_object.name,
                case_object.docs
            )

            try:
                state = CONST.TEXT.TEXT_END
                case_start_time = time.strftime("%H:%M:%S")
                case_start_timestamp = int(time.time())

                # 遍历用例的步骤
                for step in case_object.steps:
                    
                    # 替换步骤参数中的参数定义
                    step._parameters = _replace_variables_in_parameters(
                        self._task.namespace,
                        case_object.namespace,
                        step.parameters
                    )

                    try:
                        register = self._engine_action_map[step._function_name](api_engine, *step.parameters)
                    
                    except AwakenTaskCaseAssertError as error:
                        state = CONST.TEXT.TEXT_ERROR

                    except BaseException as error:
                        state = CONST.TEXT.TEXT_ERROR
                        error_message = LANGUAGE.ERROR.ERROR_MESSAGE_ENGINE_FUNCTION_RUNNING_ERROR.replace('FUNCTION_NAME', step._function_name).replace('ERROR', str(error))
                        raise AwakenTaskRunningError(error_message)

                    # 如果步骤对象存在赋值变量则赋值到对应的命名空间中
                    if step.assignment:
                        assignment_dollar_ident_count = step.assignment.count('$')
                        assignment_name = re.findall(r'\{(.*)\}', step.assignment)[0]
                        assignment_object = KeyValueObject()
                        assignment_object._number = 0
                        assignment_object._key = assignment_name
                        assignment_object._value = register

                        # 根据参数类型赋值到对应的命名空间中
                        if assignment_dollar_ident_count == 1:
                            case_object._namespace.update({assignment_object.key: assignment_object})
                        if assignment_dollar_ident_count == 2:
                            self._task._namespace.update({assignment_object.key: assignment_object})

            except BaseException as error:
                task_state = CONST.TEXT.TEXT_ERROR
                LOG.info(f'用例 [{ case_object.name }] 执行异常, 原因: { error }')
            
            finally:
                self._task_db_broker.update_case(
                    case_id,
                    state,
                    case_start_time,
                    case_start_timestamp,
                    ''  #执行步骤暂时省略
                )

        # 销毁
        self._task_db_broker.task_end(task_state)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 任务数据库代理 ]

"""
import time

from .db import DB
from .db import SQL
from ..baseic.const import CONST


class TaskDbBroker: ...


class TaskDbBroker:
    """
    [ 任务数据库代理 ]

    ---
    描述:
        该类将代理任务执行过程中的数据库写入事务。

    """

    _task_type: str
    """ 任务类型 """
    _task_id: int
    """ 任务ID """
    _task_created_date: str
    """ 任务创建日期 """
    _task_start_time: str
    """ 任务开始时间 """
    _task_start_timestamp: int
    """ 任务开始时间戳 """
    _record_cases_all: int
    """ 统计所有用例数 """
    _record_cases_success: int
    """ 统计成功用例数 """
    _record_cases_unsuccess: int
    """ 统计失败用例数 """


    def __init__(self, task_type: str):
        """
        [ 任务外交代理 ]
        
        ---
        参数:
            task_type { str } : 任务类型
        
        """
        self._record_cases_all = 0
        self._record_cases_success = 0
        self._record_cases_unsuccess = 0
        self._task_type = task_type


    def task_start(self, name: str, remark: str) -> int:
        """
        [ 任务开始 ]

        ---
        参数:
            name   { str } : 任务名称
            remark { str } : 任务注释

        ---
        返回:
            int : 创建的任务数据库记录ID
        
        """
        # 记录任务开始时间
        self._task_created_date = time.strftime('%Y-%m-%d')
        self._task_start_time = time.strftime('%H:%M:%S')
        self._task_start_timestamp = int(time.time())

        # 录入数据库任务记录
        self._task_id = DB.write(SQL.Task.creation(
            self._task_type,
            name,
            remark, 
            self._task_created_date,
        ))

        return self._task_id


    def created_case(self, name: str, remark: str) -> int:
        """
        [ 创建用例 ]

        ---
        参数:
            name   { str } : 用例名称
            remark { str } : 用例注释

        ---
        返回:
            int : 创建的用例数据库记录ID

        """
        # 录入数据库用例记录
        case_id = DB.write(SQL.Case.creation(
            self._task_id,
            name,
            self._task_type,
            remark,
            self._task_created_date
        ))
        self._record_cases_all += 1

        return case_id


    def update_case(
        self, 
        cid: int, 
        state: str, 
        start_time: str, 
        start_timestamp: int, 
        run_step: str
    ):
        """
        [ 更新用例 ]
        
        ---
        参数:
            cid             { int } : 用例ID
            state           { str } : 用例状态
            start_time      { str } : 开始时间
            start_timestamp { int } : 开始时间戳
            run_step        { str } : 执行步骤

        """
        # 录入数据库更新用例
        DB.write(SQL.Case.update(
            cid,
            state,
            start_time,
            time.strftime('%H:%M:%S'),
            (int(time.time()) - start_timestamp),
            run_step
        ))

        if state == CONST.State.Result.Success:
            self._record_cases_success += 1

        if state == CONST.State.Result.Unsuccess:
            self._record_cases_unsuccess += 1

        if state == CONST.State.Result.Error:
            self._record_cases_unsuccess += 1


    def task_end(self, state: str):
        """
        [ 任务结束 ]

        ---
        参数:
            state { str } : 任务状态

        """
        # 更新数据库任务记录
        DB.write(SQL.Task.update(
            self._task_id,
            state,
            self._task_start_time,
            time.strftime('%H:%M:%S'),
            (int(time.time()) - self._task_start_timestamp),
            self._record_cases_all,
            self._record_cases_success,
            self._record_cases_unsuccess
        ))

        # 保存事务
        DB.commit()



# Copyright 2022. quinn.7@foxmail.com All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
@ 模块     : WEB驱动引擎
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import time
import asyncio
from typing import Any

from pathlib import Path

from ...baseic.const import CONST
from ...baseic.error import AwakenWebEngineError


class WebEngine: ...


class WebEngine:
    """
    [ WEB驱动引擎 ]

    ---
    描述:
        NULL

    """

    _playwright: Any
    """ playwright 对象 """
    _browser_map: dict
    """ 浏览器类型映射字典 """
    _browser = None
    """ 浏览器进程 """
    _context = None
    """ 浏览器上下文 """
    _page = None
    """ 浏览器页面 """


    def _init_playwright(self, playwright) -> None:
        self._playwright = playwright
        self._browser_map = {
            CONST.Type.Browser.Webkit : self._playwright.webkit,
            CONST.Type.Browser.Firefox : self._playwright.firefox,
            CONST.Type.Browser.Chromium : self._playwright.chromium,
        }


    async def _browser_start(
        self, 
        browser_type=CONST.Type.Browser.Chromium, 
        headless=False
    ):
        """ 
        [ 启动浏览器 ]

        ---
        描述:
            根据传入的浏览器类型启动对象浏览器进程。

        ---
        参数:
            browser_type { str } : 浏览器类型。
            headless    { bool } : 是否以无头模式运行。

        """
        # 如果浏览器类型不存在, 则使用默认的 Chromium 浏览器
        if browser_type not in self._browser_map.keys():
            browser_type = CONST.Type.Browser.Chromium

        # 创建浏览器实例不存在则创建
        if not self._browser:
            self._browser = await self._browser_map[browser_type].launch(headless=headless)
            self._context = await self._browser.new_context()


    async def _page_start(self):
        """
        [ 启动页面 ]

        ---
        描述:
            浏览器进程启动页面实例。

        """
        self._page = await self._context.new_page()


    async def _page_close(self):
        """
        [ 关闭页面 ]
    
        ---
        描述:
            浏览器进程关闭页面实例。
        
        """
        await self._page.close()
        self._page = None


    async def _exit(self):
        """
        [ 退出浏览器 ]

        ---
        描述:
            关闭页面、上下文、浏览器对象。
        
        """
        if self._page:
            await self._page.close()
            await self._context.close()
            await self._browser.close()
        self._page = None
        self._context = None
        self._browser = None


    async def new_context(self):
        """
        [ 重置上下文 ]

        ---
        描述:
            重置浏览器的上下文, 然后重新打开一个新页面。
        
        """
        if self._page:
            await self._page.close()
            self._page = None

        await self._context.close()
        self._context = await self._browser.new_context()
        self._page = await self._context.new_page()


    async def get_page_name(self) -> str:
        """
        [ 获取页面名称 ]

        ---
        描述:
            获取当前页面中的 Title 属性。

        ---
        返回:
            str : 当前页面的名称

        """
        return await self._page.title()


    async def get_element_count(self, element: str, eid=1) -> int:
        """
        [ 获取元素数量 ]

        ---
        描述:
            获取当前页面中能够定位到的元素数量。

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        ---
        返回:
            int : 元素数量

        """
        _, count = await self._locator_elements(element, eid)
        return count


    async def get_element_text(self, element: str, eid=1) -> str:
        """
        [ 获取元素文本 ]

        ---
        描述:
            获取当前页面中定位元素的文本。

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        ---
        返回:
            str : 元素的文本

        """
        element, _ = await self._locator_elements(element, eid)
        element_text = await element.text_content()
        return element_text.strip()

    
    async def get_input_value(self, element: str, eid=1) -> str:
        """
        [ 获取输入框值 ]

        ---
        描述:
            NULL

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        ---
        返回:
            str : 元素的文本

        """
        element, _ = await self._locator_elements(element, eid)
        element_text = await element.input_value()
        return element_text

    
    async def get_attribute(self, name: str, element: str, eid=1) -> str | int:
        """
        [ 获取元素属性值 ]

        ---
        描述:
            NULL

        ---
        参数:
            name    { str } : 需获取值的属性名称
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        ---
        返回:
            str | int : 元素的属性值

        """
        element, _ = await self._locator_elements(element, eid)
        return await element.get_attribute(name)

    
    async def set_window_size(self, width: int, height: int):
        """
        [ 设置浏览器尺寸 ]

        ---
        描述:
            NULL

        ---
        参数:
            width { int }  : 宽度
            height { int } : 长度

        """
        size = {
            'width' : int(width),
            'height': int(height)
        }
        await self._page.set_viewport_size(size)


    async def goto(self, url: str):
        """
        [ 页面跳转 ]

        ---
        描述:
            NULL

        ---
        参数:
            url { str } : 跳转目标 URL 地址

        """
        await self._page.goto(url)


    async def back(self):
        """
        [ 页面后退 ]

        ---
        描述:
            NULL

        """
        await self._page.go_back()


    async def forward(self):
        """
        [ 页面前进 ]

        ---
        描述:
            NULL

        """
        await self._page.go_forward()

    
    async def reload(self):
        """
        [ 页面刷新 ]

        ---
        描述:
            NULL
            
        """
        await self._page.reload()


    async def send(self, key: str, element: str, eid=1):
        """
        [ 输入框输入 ]

        ---
        描述:
            该函数期望的目标控件是 Input。

        ---
        参数:
            key     { str } : 输入文本的值
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        """
        element, _ = await self._locator_elements(element, eid)
        await element.fill(key)


    async def click(self, element: str, eid=1):
        """
        [ 鼠标单击 ]

        ---
        描述:
            NULL

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        """
        element, _ = await self._locator_elements(element, eid)
        await element.click()


    async def dbl_click(self, element, eid=1):
        """
        [ 鼠标双击 ]

        ---
        描述:
            NULL

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        """
        element, _ = await self._locator_elements(element, eid)
        await element.dblclick()


    async def drag_to(self, element1: str, element2: str, eid=1):
        """
        [ 鼠标拖拽 ]

        ---
        描述:
            模拟鼠标拖拽行为, 将元素1拖拽至元素2, 并且松开鼠标。

        ---
        参数:
            element1 { str } : 元素1定位描述字符
            element2 { str } : 元素2定位描述字符
            eid      { int } : 元素下标

        """
        element1, _ = await self._locator_elements(element1, eid)
        element2, _ = await self._locator_elements(element2, 1)
        await element1.drag_to(element2)


    async def hover(self, wait, element: str, eid=1):
        """
        [ 鼠标悬停 ]

        ---
        描述:
            NULL

        ---
        参数:
            wait    { int } : 悬停时间
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        """
        element, _ = await self._locator_elements(element, eid)
        await element.hover()
        await asyncio.sleep(wait)


    async def popup_ok(self, element: str, eid=1, text=''):
        """
        [ 弹窗确定 ]

        ---
        描述:
            点击一个绑定了弹窗事件的按钮元素, 然后静默确认该弹窗, 如果该弹窗存在输入框, 可通过参数 text 来填充。

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标
            text    { str } : 输入文本

        """
        # 弹窗处理函数
        async def event(dialog):
            global temporary_text_register
            temporary_text_register = dialog.message
            await dialog.accept(text)

        # 绑定弹窗处理事件
        self._page.once('dialog', event)
        element, _ = await self._locator_elements(element, eid)
        await element.click()
        return temporary_text_register


    async def popup_no(self, element: str, eid=1):
        """
        [ 弹窗取消 ]

        ---
        描述:
            点击一个绑定了弹窗事件的按钮元素, 然后静默取消该弹窗。

        ---
        参数:
            element { str } : 元素定位描述字符
            eid     { int } : 元素下标

        """
        # 弹窗处理函数
        async def event(dialog):
            global temporary_text_register
            temporary_text_register = dialog.message
            await dialog.dismiss()

        # 绑定弹窗处理事件
        self._page.once('dialog', event)
        element, _ = await self._locator_elements(element, eid)
        await element.click()
        return temporary_text_register


    async def upload(self, element: str, file_path: str, eid=1):
        """
        [ 上传文件 ]

        ---
        描述:
            点击一个绑定了文件上传事件的按钮元素, 然后静默处理该事件。

        ---
        参数:
            element   { str } : 元素定位描述字符
            file_path { str } : 文件完整路径
            eid       { int } : 元素下标

        """
        # 监听文件选择器对象
        self._page.once('filechooser', lambda _: ...)
        async with self._page.expect_file_chooser() as chooser_object:
            element, _ = await self._locator_elements(element, eid)
            await element.click()

        chooser_object = await chooser_object.value
        await chooser_object.set_files(file_path)


    async def download(self, element: str, file_path: str, eid=1):
        """
        [ 下载文件 ]

        ---
        描述:
            点击一个绑定了文件下载事件的按钮元素, 然后静默处理该事件。

        ---
        参数:
            element { str } : 元素定位描述字符
            file_path { str } : 文件保存路径
            eid { int } : 元素下标

        """
        # 监听下载对象
        self._page.once('download', lambda _: ...)
        async with self._page.expect_download() as download_object:
            element, _ = await self._locator_elements(element, eid)
            await element.click()
        download_object = await download_object.value
        await download_object.save_as(file_path)
        return file_path

    
    async def save_png(self, save_path: str):
        """
        [ 保存图片 ]

        ---
        参数:
            save_path { str } : 图片保存路径

        """
        png_bytes = await self._page.screenshot()
        suffix = '.png'
        if save_path[-4:] != suffix:
            save_path = ''.join([save_path, suffix])
        save_path = Path(save_path)
        save_path.touch(0o777)
        save_path.write_bytes(png_bytes)
        return save_path.resolve()


    async def _dispatch_event(self, element, event_name):
        await self._page.dispatch_event(element, event_name)


    async def _java_script(self, expression, args):
        return await self._page.evaluate(expression, args)


    async def _locator_elements(self, element: str, eid: int):
        """
        [ 定位元素 ]

        ---
        描述:
            定位当前页面的所有符合预期的元素对象。

        ---
        参数:
            element { str } : 元素定位方式
            eid     { int } : 元素ID

        ---
        返回:
            tuple(Any | int) : 元素对象 | 元素数量

        """
        element_count: int
        retries_number = 5  # 元素重试次数
        retries_sleep  = 1  # 元素重试间隔时间

        try:
            for _ in range(retries_number):
                result = self._page.locator(element)
                element_count = await result.count()
                if element_count < 1:
                    time.sleep(retries_sleep)
                    continue
                else:
                    break
                
            if element_count == 0:
                raise BaseException

        except BaseException:
            raise AwakenWebEngineError(f'无法定位到元素 :: { element } !')
                
        try:
            # 如果 eid 大于元素数量则抛出异常
            if element_count < eid: 
                raise BaseException
            else:
                return result.nth(eid - 1), element_count

        except BaseException:
            raise AwakenWebEngineError(f'EID 超出元素对象数量, 无法应用 !')



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::指令集
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import os

from ..awaken_info import ProjectInfo
from ..baseic.common import window_template_output


def instruction_order():
    """
    [ 命令行指令::指令集 ]

    """
    window_template_output(
        f'Awaken2 - {ProjectInfo.Version}', 
        [
            '- 指令集 -',
            None,
            '-init    |  -i  |  工程初始化',
            '-make    |  -m  |  创建项目',
            '-task    |  -t  |  创建任务',
            '-run     |  -r  |  运行任务',
            '-server  |  -s  |  运行服务器',
            '-help    |  -h  |  查看帮助',
            '-version |  -v  |  查看版本',
            '-update  |  -u  |  更新版本',
        ]
    )
    exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 常量索引
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from . import const_api
from . import const_common
from . import const_error
from . import const_interpreter
from . import const_path
from . import const_state
from . import const_type


class AwakenConst: ...


class AwakenConst:
    """
    [ 常量索引 ]

    ---
    描述:
        NULL

    """


    @property
    def Api(self) -> const_api:
        """
        [ 接口常量 ]

        ---
        描述: 
            NULL

        """
        return const_api


    @property
    def Common(self) -> const_common:
        """
        [ 通用常量 ]

        ---
        描述: 
            NULL

        """
        return const_common


    @property
    def Error(self) -> const_error:
        """
        [ 错误常量 ]

        ---
        描述: 
            NULL

        """
        return const_error


    @property
    def Interpreter(self) -> const_interpreter:
        """
        [ 解释器常量 ]

        ---
        描述: 
            NULL

        """
        return const_interpreter


    @property
    def State(self) -> const_state:
        """
        [ 状态常量 ]

        ---
        描述: 
            NULL

        """
        return const_state


    @property
    def Path(self) -> const_path:
        """
        [ 路径常量 ]

        ---
        描述: 
            NULL

        """
        return const_path


    @property
    def Type(self) -> const_type:
        """
        [ 类型常量 ]

        ---
        描述: 
            NULL

        """
        return const_type


CONST: AwakenConst = AwakenConst()
""" 常量索引实例 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : Setup
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    PIP package manager build script.

"""
from setuptools import setup
from setuptools import find_packages
from src.awaken2.awaken_info import ProjectInfo


# ----------------------------------------------------------------------------
# 项目属性
# ----------------------------------------------------------------------------
NAME        = ProjectInfo.Name
VERSION     = ProjectInfo.Version
AUTHOR      = ProjectInfo.Author
EMAIL       = ProjectInfo.Email
DESCRIPTION = ProjectInfo.Desc
KEYWORDS    = ProjectInfo.Keywords
URL         = ProjectInfo.Url

LEGAL_PY_VERSION = [
    'Programming Language :: Python :: 3.10',
]

with open('README.md', 'r', encoding='UTF-8') as file:
    LONG_DESCRIPTION = file.read()

INSTALL_REQUIRES = [
    'playwright',
    'flask',
]


# ----------------------------------------------------------------------------
# 项目配置
# ----------------------------------------------------------------------------
setup(
    name                          = NAME, 
    version                       = VERSION, 
    author                        = AUTHOR, 
    author_email                  = EMAIL, 
    description                   = DESCRIPTION,
    keywords                      = KEYWORDS,
    url                           = URL,
    classifiers                   = LEGAL_PY_VERSION,
    long_description              = LONG_DESCRIPTION, 
    long_description_content_type = 'text/markdown',
    install_requires              = INSTALL_REQUIRES,
    packages                      = find_packages('src'), 
    package_dir                   = {'' : 'src'},

    entry_points = {
        'console_scripts':[
            'awaken = awaken2.awaken_console:main',
        ],
    },
)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 任务相关SQL::任务相关 ]

"""
from ..common import get_desc_limit
from ..common import parse_task_search_criteria
from ....baseic.const import CONST
from ....baseic.keyword import KEYWORD


class SqlTask(object):
    """
    [ 任务相关SQL::任务相关 ]

    """


    @staticmethod
    def get_count(paras: dict = {}) -> str:
        """
        [ 查询任务数量 ]

        ---
        可选参数:
            type                { str } : 类型
            state               { str } : 状态
            created_date        { str } : 创建日期, 例如 2022-02-02
            record_cases_state  { int } : 统计用例状态, 0表示全部成功, 1表示存在失败, 2表示存在异常

        """
        # 解析搜索条件
        where_string = parse_task_search_criteria(paras)

        return f"""

        SELECT COUNT({ KEYWORD.Api.Task.Tid })

        FROM { KEYWORD.DataBase.TableName.Tasks }

        {where_string}

        """


    @staticmethod
    def get_id_list(paras: dict = {}) -> str:
        """
        [ 查询 ID 列表 ]

        ---
        可选参数:
            type                { str } : 类型
            state               { str } : 状态
            created_date        { str } : 创建日期, 例如 2022-02-02
            record_cases_state  { int } : 统计用例状态, 0表示全部成功, 1表示存在失败, 2表示存在异常

        """
        # 解析搜索条件
        where_string = parse_task_search_criteria(paras)

        return f"""

        SELECT { KEYWORD.Api.Task.Tid }

        FROM { KEYWORD.DataBase.TableName.Tasks }

        {where_string}

        ORDER BY { KEYWORD.Api.Task.Tid } DESC

        """


    @staticmethod
    def get_exist_task_dates(paras: dict = {}) -> str:
        """
        [ 查询去重后的所有存在任务的日期列表 ]

        ---
        可选参数:
            type                { str } : 类型
            state               { str } : 状态
            created_date        { str } : 创建日期, 例如 2022-02-02
            record_cases_state  { int } : 统计用例状态, 0表示全部成功, 1表示存在失败, 2表示存在异常

        """
        # 解析搜索条件
        where_string = parse_task_search_criteria(paras)

        return f"""

        SELECT DISTINCT { KEYWORD.Api.Task.CreatedDate } 

        FROM { KEYWORD.DataBase.TableName.Tasks }

        {where_string}

        ORDER BY { KEYWORD.Api.Task.Tid } DESC

        """


    @staticmethod
    def get_spend_time(paras: dict = {}) -> str:
        """
        [ 查询任务的消耗时间 ]

        ---
        可选参数:
            tid                 { int } : 任务编号
            type                { str } : 类型
            state               { str } : 状态
            created_date        { str } : 创建日期, 例如 2022-02-02
            record_cases_state  { int } : 统计用例状态, 0表示全部成功, 1表示存在失败, 2表示存在异常

        """
        # 解析搜索条件
        where_string = parse_task_search_criteria(paras)

        return f""" 

        SELECT SUM({ KEYWORD.Api.Task.SpendTime })

        FROM { KEYWORD.DataBase.TableName.Tasks }

        {where_string}

        """


    @staticmethod
    def get_task_list(paras: dict = {}) -> str:
        """
        [ 查询任务列表 ]

        ---
        可选参数:
            tid                 { int } : 任务编号
            type                { str } : 类型
            state               { str } : 状态
            created_date        { str } : 创建日期, 例如 2022-02-02
            record_cases_state  { int } : 统计用例状态, 0表示全部成功, 1表示存在失败, 2表示存在异常
            page                { int } : 页数
            number              { int } : 单页数据量

        """
        # 解析搜索条件
        where_string = parse_task_search_criteria(paras)

        limit = ''
        if 'page' in paras.keys() and 'number' in paras.keys():
            limit = f"LIMIT { get_desc_limit(paras['page'], paras['number']) }"

        return f"""

        SELECT * 

        FROM { KEYWORD.DataBase.TableName.Tasks }

        {where_string}

        ORDER BY { KEYWORD.Api.Task.Tid } DESC

        {limit}

        """


    @staticmethod
    def creation(
        type: str,
        name: str,
        remark: str,
        created_date: str
    ) -> str:
        """
        [ 创建任务 ]

        ---
        参数:
            type         { str } : 任务类型
            name         { str } : 任务名称
            remark       { str } : 任务备注
            created_date { str } : 创建日期, 例如 2022-02-02

        """
        return f""" 

        INSERT INTO { KEYWORD.DataBase.TableName.Tasks } (

            { KEYWORD.Api.Task.Tid },
            { KEYWORD.Api.Task.Type },
            { KEYWORD.Api.Task.Name },
            { KEYWORD.Api.Task.Remark }, 
            { KEYWORD.Api.Task.State }, 
            { KEYWORD.Api.Task.CreatedDate }, 
            { KEYWORD.Api.Task.StartTime }, 
            { KEYWORD.Api.Task.EndTime }, 
            { KEYWORD.Api.Task.SpendTime }, 
            { KEYWORD.Api.Task.RecordCasesAll }, 
            { KEYWORD.Api.Task.RecordCasesSuccess }, 
            { KEYWORD.Api.Task.RecordCasesUnSuccess }
        )

        VALUES (

            NULL,
            '{ type }',
            '{ name }',
            '{ remark }',
            '{ CONST.State.Result.Null }',
            '{ created_date }',
            NULL, NULL, NULL, NULL, NULL, NULL
        )

        """


    @staticmethod
    def update(
        tid: int,
        state: str,
        start_time: str,
        end_time: str,
        spend_time: int,
        record_cases_all: int,
        record_cases_success: int,
        record_cases_unsuccess: int
    ) -> str:
        """
        [ 更新任务 ]

        ---
        参数:
            tid                    { int } : 任务编号
            state                  { str } : 任务状态
            start_time             { str } : 开始时间
            end_time               { str } : 结束时间
            spend_time             { int } : 消耗时间
            record_cases_all       { int } : 统计所有用例数
            record_cases_success   { int } : 统计成功用例数
            record_cases_unsuccess { int } : 统计成功用例数

        """
        return f"""

        UPDATE { KEYWORD.DataBase.TableName.Tasks }

        SET
            { KEYWORD.Api.Task.State }                 = '{ state }', 
            { KEYWORD.Api.Task.StartTime }             = '{ start_time }', 
            { KEYWORD.Api.Task.EndTime }               = '{ end_time }', 
            { KEYWORD.Api.Task.SpendTime }             = '{ spend_time }', 
            { KEYWORD.Api.Task.RecordCasesAll }        =  { record_cases_all }, 
            { KEYWORD.Api.Task.RecordCasesSuccess }    =  { record_cases_success }, 
            { KEYWORD.Api.Task.RecordCasesUnSuccess }  =  { record_cases_unsuccess }

        WHERE 
            { KEYWORD.Api.Task.Tid } = { tid }

        """


    @staticmethod
    def delete(paras: dict = {}) -> str:
        """
        [ 删除任务 ]

        ---
        可选参数:
            tid                 { int } : 任务编号
            type                { str } : 类型
            state               { str } : 状态
            created_date        { str } : 创建日期, 例如 2022-02-02
            record_cases_state  { int } : 统计用例状态, 0表示全部成功, 1表示存在失败, 2表示存在异常

        """
        # 解析搜索条件
        where_string = parse_task_search_criteria(paras)

        return f"""

        DELETE FROM { KEYWORD.DataBase.TableName.Tasks }

        {where_string}

        """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ Awaken 数据库 ]

"""
import sqlite3
from threading import Lock

from ...baseic.const import CONST
from ...baseic.keyword import KEYWORD
from ...baseic.error import AwakenDataBaseError
from ...baseic.decorator import singleton_pattern


@singleton_pattern
class AwakenDb(object):
    """
    [ Awaken 数据库 ]

    ---
    描述:
        基于 SQLite 数据库。

    """
    
    _conn: sqlite3.Connection = None
    """ 数据库连接对象 """
    _lock: Lock = Lock()
    """ 数据库互斥锁 """


    def __init__(self):
        """
        [ Rains 数据库 ]

        ---
        异常:
            AwakenDbInitError : Awaken 数据库初始化异常。

        """
        try:
            # 创建/连接数据库
            if not self._conn:
                self.connect()
                self._conn.commit()

        except BaseException as error:
            raise AwakenDataBaseError(error)


    def read(self, sql: str) -> list:
        """
        [ 无锁读取 ]

        ---
        描述:
            执行查询语句, 该函数不会持有数据库锁。

        ---
        参数:
            sql { str } : SQLite 查询语句。

        ---
        异常:
            AwakenDbReadError : Awaken 数据库读取异常。

        ---
        返回:
            list : 查询结果列表。

        """
        try:
            # 创建数据库游标
            cur = self._conn.cursor()
            # 执行查询语句
            cur.execute(sql)
            # 获取查询结果
            data = cur.fetchall()
            # 注销数据库游标
            cur.close()
            return data

        except BaseException as error:
            raise AwakenDataBaseError(error)


    def lock_read(self, sql: str) -> list:
        """
        [ 持锁读取 ]

        ---
        参数:
            执行查询语句, 该函数会持有数据库锁, 结束时释放。

        ---
        参数:
            sql { str } : SQLite 查询语句。

        ---
        异常:
            AwakenDbReadError : Awaken 数据库读取异常。

        ---
        返回:
            list : 查询结果列表。

        """
        # 获取锁
        with self._lock:

            # 执行读取函数
            return self.read(sql)


    def write(self, sql: str) -> int:
        """
        [ 无锁写入 ]

        描述:
            执行写入语句, 该函数不会持有数据库锁。

        ---
        参数:
            sql { str } : SQLite 查询语句。

        ---
        异常:
            AwakenDbWriteError : Awaken 数据库读取异常。

        ---
        返回:
            int : 写入数据后返回该条数据的自增ID。

        """
        try:
            # 创建数据库游标
            cur = self._conn.cursor()
            # 执行查询语句
            cur.execute(sql)
            # 获取数据自增ID
            aid = cur.lastrowid
            # 注销数据库游标
            cur.close()
            return aid

        except BaseException as error:
            raise AwakenDataBaseError(error)


    def lock_write(self, sql: str) -> int:
        """
        [ 持锁写入 ]

        ---
        描述:
            执行写入语句, 该函数会持有数据库锁, 结束时释放。

        ---
        参数:
            sql { str } : SQLite 查询语句。

        ---
        异常:
            AwakenDbWriteError : Awaken 数据库读取异常。

        ---
        返回:
            int : 写入数据后返回该条数据的自增ID。

        """
        # 获取锁
        with self._lock:

            # 执行写入函数
            return self.write(sql)


    def connect(self):
        """
        [ 数据库连接 ]

        """

        # 创建数据库连接对象
        self._conn = sqlite3.connect(CONST.Path.FilePath.Database, check_same_thread=False)

        # 创建数据库游标
        Cur = self._conn.cursor()

        # 构建数据库表
        for key, value in CreateTableSql.__dict__.items():
            if '__' in key:
                continue
            Cur.execute(value)

        # 注销数据库游标
        Cur.close()


    def commit(self):
        """
        [ 事务保存 ]

        """
        self._conn.commit()


    def rollback(self):
        """
        [ 事务回滚 ]

        """
        self._conn.rollback()


    def quit(self):
        """
        [ 关闭数据库连接 ]

        """
        self._conn.close()


class CreateTableSql(object):
    """
    [ 建表语句 ]

    """

    Tasks = f""" 

    CREATE TABLE IF NOT EXISTS { KEYWORD.DataBase.TableName.Tasks }
    (
        { KEYWORD.Api.Task.Tid }                   INTEGER PRIMARY KEY,
        { KEYWORD.Api.Task.Type }                  TEXT NOT NULL,
        { KEYWORD.Api.Task.Name }                  TEXT NOT NULL,
        { KEYWORD.Api.Task.Remark }                TEXT NOT NULL,
        { KEYWORD.Api.Task.State }                 TEXT NOT NULL,
        { KEYWORD.Api.Task.CreatedDate }           DATE NOT NULL,
        { KEYWORD.Api.Task.StartTime }             DATE,
        { KEYWORD.Api.Task.EndTime }               DATE,
        { KEYWORD.Api.Task.SpendTime }             INT,
        { KEYWORD.Api.Task.RecordCasesAll }        INT,
        { KEYWORD.Api.Task.RecordCasesSuccess }    INT,
        { KEYWORD.Api.Task.RecordCasesUnSuccess }  INT
    );

    """
    """
    [ 建表语句::任务表 ]

    """

    Cases = f"""
    
    CREATE TABLE IF NOT EXISTS { KEYWORD.DataBase.TableName.Cases }
    (
        { KEYWORD.Api.Case.Cid }          INTEGER PRIMARY KEY, 
        { KEYWORD.Api.Case.Tid }          INT  NOT NULL, 
        { KEYWORD.Api.Case.Name }         TEXT NOT NULL,
        { KEYWORD.Api.Case.Type }         TEXT NOT NULL, 
        { KEYWORD.Api.Case.Remark }       TEXT NOT NULL, 
        { KEYWORD.Api.Case.State }        TEXT NOT NULL, 
        { KEYWORD.Api.Case.CreatedDate }  DATE NOT NULL, 
        { KEYWORD.Api.Case.StartTime }    DATE,
        { KEYWORD.Api.Case.EndTime }      DATE,
        { KEYWORD.Api.Case.SpendTime }    INT,
        { KEYWORD.Api.Case.RunStep }      TEXT
    );

    """
    """
    [ 建表语句::用例表 ]

    """ 
